@@ -15,8 +15,8 @@
  *
  * You should have received a copy of the GNU Library General Public
  * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
+ * Boston, MA 02110-1301, USA.
  */
 
 /**
@@ -62,9 +62,10 @@
  *     </para></listitem>
  *     <listitem><para>
  *       If codec processing results in encoded data, subclass should call
- *       @gst_audio_encoder_finish_frame to have encoded data pushed
- *       downstream.  Alternatively, it might also call to indicate dropped
- *       (non-encoded) samples.
+ *       gst_audio_encoder_finish_frame() to have encoded data pushed
+ *       downstream. Alternatively, it might also call
+ *       gst_audio_encoder_finish_frame() (with a NULL buffer and some number of
+ *       dropped samples) to indicate dropped (non-encoded) samples.
  *     </para></listitem>
  *     <listitem><para>
  *       Just prior to actually pushing a buffer downstream,
@@ -140,7 +141,7 @@
  *   </para></listitem>
  *   <listitem><para>
  *      Accept data in @handle_frame and provide encoded results to
- *      @gst_audio_encoder_finish_frame.
+ *      gst_audio_encoder_finish_frame().
  *   </para></listitem>
  * </itemizedlist>
  *
@@ -185,6 +186,9 @@ enum
 typedef struct _GstAudioEncoderContext
 {
   /* input */
+  /* last negotiated input caps */
+  GstCaps *input_caps;
+  /* last negotiated input info */
   GstAudioInfo info;
 
   /* output */
@@ -461,7 +465,13 @@ gst_audio_encoder_reset (GstAudioEncoder * enc, gboolean full)
     enc->priv->ctx.headers = NULL;
     enc->priv->ctx.new_headers = FALSE;
 
+    if (enc->priv->ctx.allocator)
+      gst_object_unref (enc->priv->ctx.allocator);
+    enc->priv->ctx.allocator = NULL;
+
+    gst_caps_replace (&enc->priv->ctx.input_caps, NULL);
     gst_caps_replace (&enc->priv->ctx.caps, NULL);
+
     memset (&enc->priv->ctx, 0, sizeof (enc->priv->ctx));
     gst_audio_info_init (&enc->priv->ctx.info);
 
@@ -473,10 +483,6 @@ gst_audio_encoder_reset (GstAudioEncoder * enc, gboolean full)
     g_list_foreach (enc->priv->pending_events, (GFunc) gst_event_unref, NULL);
     g_list_free (enc->priv->pending_events);
     enc->priv->pending_events = NULL;
-
-    if (enc->priv->ctx.allocator)
-      gst_object_unref (enc->priv->ctx.allocator);
-    enc->priv->ctx.allocator = NULL;
   }
 
   gst_segment_init (&enc->input_segment, GST_FORMAT_TIME);
@@ -621,7 +627,11 @@ gst_audio_encoder_finish_frame (GstAudioEncoder * enc, GstBuffer * buf,
   if (G_UNLIKELY (ctx->output_caps_changed
           || gst_pad_check_reconfigure (enc->srcpad))) {
     if (!gst_audio_encoder_negotiate (enc)) {
-      ret = GST_FLOW_NOT_NEGOTIATED;
+      gst_pad_mark_reconfigure (enc->srcpad);
+      if (GST_PAD_IS_FLUSHING (enc->srcpad))
+        ret = GST_FLOW_FLUSHING;
+      else
+        ret = GST_FLOW_NOT_NEGOTIATED;
       goto exit;
     }
   }
@@ -674,7 +684,7 @@ gst_audio_encoder_finish_frame (GstAudioEncoder * enc, GstBuffer * buf,
     if (!enc->priv->perfect_ts) {
       guint64 ts, distance;
 
-      ts = gst_adapter_prev_timestamp (priv->adapter, &distance);
+      ts = gst_adapter_prev_pts (priv->adapter, &distance);
       g_assert (distance % ctx->info.bpf == 0);
       distance /= ctx->info.bpf;
       GST_LOG_OBJECT (enc, "%" G_GUINT64_FORMAT " samples past prev_ts %"
@@ -1094,7 +1104,6 @@ gst_audio_encoder_chain (GstPad * pad, GstObject * parent, GstBuffer * buffer)
   }
 
   /* clip to segment */
-  /* NOTE: slightly painful linking -laudio only for this one ... */
   buffer = gst_audio_buffer_clip (buffer, &enc->input_segment, ctx->info.rate,
       ctx->info.bpf);
   if (G_UNLIKELY (!buffer)) {
@@ -1212,33 +1221,16 @@ wrong_buffer:
   }
 }
 
-static gboolean
-audio_info_is_equal (GstAudioInfo * from, GstAudioInfo * to)
-{
-  if (from == to)
-    return TRUE;
-  if (from->finfo == NULL || to->finfo == NULL)
-    return FALSE;
-  if (GST_AUDIO_INFO_FORMAT (from) != GST_AUDIO_INFO_FORMAT (to))
-    return FALSE;
-  if (GST_AUDIO_INFO_RATE (from) != GST_AUDIO_INFO_RATE (to))
-    return FALSE;
-  if (GST_AUDIO_INFO_CHANNELS (from) != GST_AUDIO_INFO_CHANNELS (to))
-    return FALSE;
-  if (GST_AUDIO_INFO_CHANNELS (from) > 64)
-    return TRUE;
-  return (memcmp (from->position, to->position,
-          GST_AUDIO_INFO_CHANNELS (from) * sizeof (to->position[0])) == 0);
-}
-
 static gboolean
 gst_audio_encoder_sink_setcaps (GstAudioEncoder * enc, GstCaps * caps)
 {
   GstAudioEncoderClass *klass;
   GstAudioEncoderContext *ctx;
   GstAudioInfo state;
-  gboolean res = TRUE, changed = FALSE;
+  gboolean res = TRUE;
   guint old_rate;
+  GstClockTime old_min_latency;
+  GstClockTime old_max_latency;
 
   klass = GST_AUDIO_ENCODER_GET_CLASS (enc);
 
@@ -1254,6 +1246,16 @@ gst_audio_encoder_sink_setcaps (GstAudioEncoder * enc, GstCaps * caps)
   if (!gst_caps_is_fixed (caps))
     goto refuse_caps;
 
+  if (enc->priv->ctx.input_caps
+      && gst_caps_is_equal (enc->priv->ctx.input_caps, caps))
+    goto same_caps;
+
+  if (!gst_audio_info_from_caps (&state, caps))
+    goto refuse_caps;
+
+  if (enc->priv->ctx.input_caps && gst_audio_info_is_equal (&state, &ctx->info))
+    goto same_caps;
+
   /* adjust ts tracking to new sample rate */
   old_rate = GST_AUDIO_INFO_RATE (&ctx->info);
   if (GST_CLOCK_TIME_IS_VALID (enc->priv->base_ts) && old_rate) {
@@ -1262,64 +1264,58 @@ gst_audio_encoder_sink_setcaps (GstAudioEncoder * enc, GstCaps * caps)
     enc->priv->samples = 0;
   }
 
-  if (!gst_audio_info_from_caps (&state, caps))
-    goto refuse_caps;
+  /* drain any pending old data stuff */
+  gst_audio_encoder_drain (enc);
 
-  changed = !audio_info_is_equal (&state, &ctx->info);
+  /* context defaults */
+  enc->priv->ctx.frame_samples_min = 0;
+  enc->priv->ctx.frame_samples_max = 0;
+  enc->priv->ctx.frame_max = 0;
+  enc->priv->ctx.lookahead = 0;
 
-  if (changed) {
-    GstClockTime old_min_latency;
-    GstClockTime old_max_latency;
+  /* element might report latency */
+  GST_OBJECT_LOCK (enc);
+  old_min_latency = ctx->min_latency;
+  old_max_latency = ctx->max_latency;
+  GST_OBJECT_UNLOCK (enc);
 
-    /* drain any pending old data stuff */
-    gst_audio_encoder_drain (enc);
+  if (klass->set_format)
+    res = klass->set_format (enc, &state);
 
-    /* context defaults */
-    enc->priv->ctx.frame_samples_min = 0;
-    enc->priv->ctx.frame_samples_max = 0;
-    enc->priv->ctx.frame_max = 0;
-    enc->priv->ctx.lookahead = 0;
-
-    /* element might report latency */
-    GST_OBJECT_LOCK (enc);
-    old_min_latency = ctx->min_latency;
-    old_max_latency = ctx->max_latency;
-    GST_OBJECT_UNLOCK (enc);
-
-    if (klass->set_format)
-      res = klass->set_format (enc, &state);
-
-    if (res)
-      ctx->info = state;
-
-    /* invalidate state to ensure no casual carrying on */
-    if (!res) {
-      GST_DEBUG_OBJECT (enc, "subclass did not accept format");
-      gst_audio_info_init (&state);
-      goto exit;
-    }
-
-    /* notify if new latency */
-    GST_OBJECT_LOCK (enc);
-    if ((ctx->min_latency > 0 && ctx->min_latency != old_min_latency) ||
-        (ctx->max_latency > 0 && ctx->max_latency != old_max_latency)) {
-      GST_OBJECT_UNLOCK (enc);
-      /* post latency message on the bus */
-      gst_element_post_message (GST_ELEMENT (enc),
-          gst_message_new_latency (GST_OBJECT (enc)));
-      GST_OBJECT_LOCK (enc);
-    }
-    GST_OBJECT_UNLOCK (enc);
+  if (res) {
+    ctx->info = state;
+    gst_caps_replace (&enc->priv->ctx.input_caps, caps);
   } else {
-    GST_DEBUG_OBJECT (enc, "new audio format identical to configured format");
+    /* invalidate state to ensure no casual carrying on */
+    GST_DEBUG_OBJECT (enc, "subclass did not accept format");
+    gst_audio_info_init (&state);
+    goto exit;
   }
 
+  /* notify if new latency */
+  GST_OBJECT_LOCK (enc);
+  if ((ctx->min_latency > 0 && ctx->min_latency != old_min_latency) ||
+      (ctx->max_latency > 0 && ctx->max_latency != old_max_latency)) {
+    GST_OBJECT_UNLOCK (enc);
+    /* post latency message on the bus */
+    gst_element_post_message (GST_ELEMENT (enc),
+        gst_message_new_latency (GST_OBJECT (enc)));
+    GST_OBJECT_LOCK (enc);
+  }
+  GST_OBJECT_UNLOCK (enc);
+
 exit:
 
   GST_AUDIO_ENCODER_STREAM_UNLOCK (enc);
 
   return res;
 
+same_caps:
+  {
+    GST_DEBUG_OBJECT (enc, "new audio format identical to configured format");
+    goto exit;
+  }
+
   /* ERRORS */
 refuse_caps:
   {
@@ -2042,6 +2038,7 @@ gst_audio_encoder_set_frame_samples_min (GstAudioEncoder * enc, gint num)
   g_return_if_fail (GST_IS_AUDIO_ENCODER (enc));
 
   enc->priv->ctx.frame_samples_min = num;
+  GST_LOG_OBJECT (enc, "set to %d", num);
 }
 
 /**
@@ -2075,6 +2072,7 @@ gst_audio_encoder_set_frame_samples_max (GstAudioEncoder * enc, gint num)
   g_return_if_fail (GST_IS_AUDIO_ENCODER (enc));
 
   enc->priv->ctx.frame_samples_max = num;
+  GST_LOG_OBJECT (enc, "set to %d", num);
 }
 
 /**
@@ -2105,6 +2103,7 @@ gst_audio_encoder_set_frame_max (GstAudioEncoder * enc, gint num)
   g_return_if_fail (GST_IS_AUDIO_ENCODER (enc));
 
   enc->priv->ctx.frame_max = num;
+  GST_LOG_OBJECT (enc, "set to %d", num);
 }
 
 /**
@@ -2134,6 +2133,7 @@ gst_audio_encoder_set_lookahead (GstAudioEncoder * enc, gint num)
   g_return_if_fail (GST_IS_AUDIO_ENCODER (enc));
 
   enc->priv->ctx.lookahead = num;
+  GST_LOG_OBJECT (enc, "set to %d", num);
 }
 
 /**
@@ -2168,6 +2168,9 @@ gst_audio_encoder_set_latency (GstAudioEncoder * enc,
   enc->priv->ctx.min_latency = min;
   enc->priv->ctx.max_latency = max;
   GST_OBJECT_UNLOCK (enc);
+
+  GST_LOG_OBJECT (enc, "set to %" GST_TIME_FORMAT "-%" GST_TIME_FORMAT,
+      GST_TIME_ARGS (min), GST_TIME_ARGS (max));
 }
 
 /**
@@ -2367,6 +2370,8 @@ gst_audio_encoder_set_tolerance (GstAudioEncoder * enc, GstClockTime tolerance)
   GST_OBJECT_LOCK (enc);
   enc->priv->tolerance = tolerance;
   GST_OBJECT_UNLOCK (enc);
+
+  GST_LOG_OBJECT (enc, "set to %" GST_TIME_FORMAT, GST_TIME_ARGS (tolerance));
 }
 
 /**
@@ -2525,11 +2530,11 @@ static gboolean
 gst_audio_encoder_negotiate_default (GstAudioEncoder * enc)
 {
   GstAudioEncoderClass *klass;
-  gboolean res = FALSE;
+  gboolean res = TRUE;
   GstQuery *query = NULL;
   GstAllocator *allocator;
   GstAllocationParams params;
-  GstCaps *caps;
+  GstCaps *caps, *prevcaps;
 
   g_return_val_if_fail (GST_IS_AUDIO_ENCODER (enc), FALSE);
   g_return_val_if_fail (GST_IS_CAPS (enc->priv->ctx.caps), FALSE);
@@ -2540,7 +2545,33 @@ gst_audio_encoder_negotiate_default (GstAudioEncoder * enc)
 
   GST_DEBUG_OBJECT (enc, "Setting srcpad caps %" GST_PTR_FORMAT, caps);
 
-  res = gst_pad_set_caps (enc->srcpad, caps);
+  if (enc->priv->pending_events) {
+    GList **pending_events, *l;
+
+    pending_events = &enc->priv->pending_events;
+
+    GST_DEBUG_OBJECT (enc, "Pushing pending events");
+    for (l = *pending_events; l;) {
+      GstEvent *event = GST_EVENT (l->data);
+      GList *tmp;
+
+      if (GST_EVENT_TYPE (event) < GST_EVENT_CAPS) {
+        gst_audio_encoder_push_event (enc, l->data);
+        tmp = l;
+        l = l->next;
+        *pending_events = g_list_delete_link (*pending_events, tmp);
+      } else {
+        l = l->next;
+      }
+    }
+  }
+
+  prevcaps = gst_pad_get_current_caps (enc->srcpad);
+  if (!prevcaps || !gst_caps_is_equal (prevcaps, caps))
+    res = gst_pad_set_caps (enc->srcpad, caps);
+  if (prevcaps)
+    gst_caps_unref (prevcaps);
+
   if (!res)
     goto done;
   enc->priv->ctx.output_caps_changed = FALSE;
@@ -2682,15 +2713,27 @@ gst_audio_encoder_allocate_output_buffer (GstAudioEncoder * enc, gsize size)
 
   if (G_UNLIKELY (enc->priv->ctx.output_caps_changed || (enc->priv->ctx.caps
               && gst_pad_check_reconfigure (enc->srcpad)))) {
-    if (!gst_audio_encoder_negotiate (enc))
-      goto done;
+    if (!gst_audio_encoder_negotiate (enc)) {
+      GST_INFO_OBJECT (enc, "Failed to negotiate, fallback allocation");
+      gst_pad_mark_reconfigure (enc->srcpad);
+      goto fallback;
+    }
   }
 
   buffer =
       gst_buffer_new_allocate (enc->priv->ctx.allocator, size,
       &enc->priv->ctx.params);
+  if (!buffer) {
+    GST_INFO_OBJECT (enc, "couldn't allocate output buffer");
+    goto fallback;
+  }
 
-done:
+  GST_AUDIO_ENCODER_STREAM_UNLOCK (enc);
+
+  return buffer;
+
+fallback:
+  buffer = gst_buffer_new_allocate (NULL, size, NULL);
   GST_AUDIO_ENCODER_STREAM_UNLOCK (enc);
 
   return buffer;
